/* -*-c-*- */
/* int CORE = 1; */
/* int FUNCTIONS = 1; */
/* int X11 = 1; */
/* int FVWM = 1; */
/* int CONFIG = 1; */
/* int WINLIST = 1; */
/* int MEM = 1; */

/* I'm finding lots of the debugging is dereferencing pointers
   to zero.  I fixed some of them, until I grew tired of the game.
   If you want to turn these back on, be prepared for lots of core
   dumps.  dje 11/15/98. */
int CORE = 0;
int FUNCTIONS = 0;
int X11 = 0;
int FVWM = 0;
int CONFIG = 0;
int WINLIST = 0;
int MEM = 0;
