/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see: <http://www.gnu.org/licenses/>
 */

#ifndef FVWMLIB_GETPWUID_H
#define FVWMLIB_GETPWUID_H

#ifdef HAVE_GETPWUID
#  include <pwd.h>
#endif

const char	*find_home_dir(void);
char	*expand_path(const char *);

#endif
